from __future__ import annotations

import asyncio
from dataclasses import dataclass
from datetime import datetime, timedelta, tzinfo
from typing import Dict, List, Optional, Any

import disnake
import pytz
from disnake.ext import commands

from .base import DatabaseChecker


__all__ = ("PartialBirthdayMember", "BirthdayManager", "BirthdayMember")


@dataclass
class PartialBirthdayMember:
    """
    Represents a partial birthday member.
    """

    member: disnake.Member
    birthday_date: datetime
    timezone: tzinfo


@dataclass
class BirthdayMember:
    """
    Represents a birthday member.
    """

    birthday_manager: BirthdayManager
    member: disnake.Member

    def __post_init__(self):
        self.table = self.birthday_manager.tables["birthdays"]

    def __hash__(self):
        return id(self)

    @property
    def __checks(self) -> Dict[str, int]:
        return {"guild": self.member.guild.id, "member": self.member.id}

    async def birthday_date(self) -> datetime:
        """
        |coro|

        Returns the birthday date in UTC of the member.

        :return: The birthday.
        :rtype: datetime
        """

        birthday_data = await self.birthday_manager.database.select(
            self.table, ["utc_birthday"], self.__checks
        )
        return datetime.fromtimestamp(
            birthday_data["utc_birthday"], await self.timezone()
        )

    async def next_birthday(self) -> datetime:
        """
        |coro|

        Returns the next birthday of the member.

        :return: The next birthday of the member.
        :rtype: datetime
        """

        current_datetime = datetime.now(await self.timezone())

        new_date = (await self.birthday_date()).replace(year=current_datetime.year)
        if new_date.timestamp() - current_datetime.timestamp() < 0:
            new_date = new_date.replace(year=current_datetime.year + 1)

        return new_date

    async def timezone(self) -> tzinfo:
        """
        |coro|

        Returns the timezone.

        :return: The timezone.
        :rtype: str
        """

        timezone_data = await self.birthday_manager.database.select(
            self.table, ["timezone"], self.__checks
        )
        return pytz.timezone(timezone_data["timezone"])

    async def delete(self) -> PartialBirthdayMember:
        """
        |coro|

        Deletes the birthday and returns a PartialBirthdayMember.

        :return: The partial birthday.
        :rtype: PartialBirthdayMember
        """

        partial = PartialBirthdayMember(
            self.member, await self.birthday_date(), await self.timezone()
        )
        await self.birthday_manager.database.delete(self.table, self.__checks)
        return partial

    async def set_birthday_date(self, timestamp: float) -> None:
        """
        |coro|

        Sets the birthday date.

        :param float timestamp: The timestamp.
        :return: None
        :rtype: None
        """

        await self.birthday_manager.database.update(
            self.table, {"utc_birthday": timestamp}, self.__checks
        )

    async def set_timezone(self, timezone: str) -> None:
        """
        |coro|

        Sets the member's timezone.

        :param str timezone: The timezone.
        :return: None
        :rtype: None
        """

        await self.birthday_manager.database.update(
            self.table, {"timezone": timezone}, self.__checks
        )

    async def age(self) -> int:
        """
        |coro|

        Returns the current age of the member.

        :return: The member age.
        :rtype: int
        """

        born = await self.birthday_date()
        today = datetime.now(await self.timezone())

        return (
            today.year - born.year - ((today.month, today.day) < (born.month, born.day))
        )


class BirthdayManager(DatabaseChecker):
    """
    Represents a birthday manager.
    """

    def __init__(self, bot: commands.Bot):
        """
        :param commands.Bot bot: The bot.
        """

        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "member": "snowflake",
                    "utc_birthday": "snowflake",
                    "timezone": "string",
                }
            ],
            ["birthdays"],
        )
        self.bot = bot
        self.add_event(self._on_database_connect, "on_database_connect")

    async def _on_database_connect(self):
        self.bot.loop.create_task(self.__detect_birthdays())

    @DatabaseChecker.uses_database
    async def create_birthday(
        self, member: disnake.Member, member_birthday: float, timezone: str = "UTC"
    ) -> None:
        """
        |coro|

        Makes a birthday for the member.

        :param disnake.Member member: The member.
        :param float member_birthday: The member birthday timestamp in UTC.
        :param str timezone: The timezone.
        :return: None
        :rtype: None
        """

        await self.database.insertifnotexists(
            self.tables["birthdays"],
            dict(
                zip(
                    self.tables_column_data[0],
                    [member.guild.id, member.id, member_birthday, timezone],
                )
            ),
            {"guild": member.guild.id, "member": member.id},
        )

    @DatabaseChecker.uses_database
    async def get_birthday(self, member: disnake.Member) -> Optional[BirthdayMember]:
        """
        |coro|

        Returns the BirthdayMember object of the member.

        :param disnake.Member member: The member.
        :return: The BirthdayMember object if applicable.
        :rtype: Optional[BirthdayMember]
        """

        member_data = await self.database.select(
            self.tables["birthdays"],
            [],
            {"guild": member.guild.id, "member": member.id},
            True,
        )

        if member_data:
            return BirthdayMember(self, member)

        return None

    @DatabaseChecker.uses_database
    async def get_upcoming(self, guild: disnake.Guild) -> List[BirthdayMember]:
        """
        |coro|

        Returns the upcoming birthdays in the guild.

        :param disnake.Guild guild: The guild.
        :return: The birthdays, sorted by their nearest birthday date.
        :rtype: List[BirthdayMember]
        """

        member_data = await self.database.select(
            self.tables["birthdays"], [], {"guild": guild.id}, fetchall=True
        )

        birthdays = {}

        for birthday_member in member_data:
            member = guild.get_member(birthday_member["member"])

            if member:
                birthday = BirthdayMember(self, member)
                birthdays[birthday] = await birthday.next_birthday()
                # Did this instead of running the method until complete in the sorted key lambda

        return sorted(birthdays, key=lambda x: birthdays[x])

    @staticmethod
    def get_midnight_timezones() -> List[str]:
        """
        This method returns a list of timezones where the current time is 12 am.

        :return: The list of timezones.
        :rtype: List[str]
        """

        current_utc_time = datetime.utcnow()
        utc_offset = -(current_utc_time.hour % 24)

        minutes = 30 if current_utc_time.minute > 5 else 0
        if minutes == 30:
            utc_offset -= 1

        checks = (
            timedelta(hours=utc_offset, minutes=minutes),
            timedelta(
                hours=24 - -utc_offset if current_utc_time.hour != 0 else 0,
                minutes=minutes,
            ),
        )

        return [
            tz.zone
            for tz in map(pytz.timezone, pytz.all_timezones_set)
            if current_utc_time.astimezone(tz).utcoffset() in checks
        ]

    @DatabaseChecker.uses_database
    async def get_members_with_birthday(
        self, timezones: List[str]
    ) -> List[Dict[str, Any]]:
        """
        |coro|

        This function receives a list of timezones and returns a list of members that have birthdays in that date
        and timezone.

        :param List[str] timezones: The timezones.
        :return: Returns the members that have a birthday.
        :rtype: List[Dict[str, Any]]
        """

        result_members = []
        registered_members = await self.database.select(
            self.tables["birthdays"], [], fetchall=True
        )

        birthday_members = [x for x in registered_members if x["timezone"] in timezones]
        for birthday_member in birthday_members:
            timezone_time = datetime.now(pytz.timezone(birthday_member["timezone"]))
            date_of_birth = datetime.fromtimestamp(birthday_member["utc_birthday"])

            if (
                date_of_birth.month == timezone_time.month
                and date_of_birth.day == timezone_time.day
            ):
                result_members.append(birthday_member)

        return result_members

    @staticmethod
    def round_to_nearest(timedelta_to_round: timedelta) -> float:
        """
        This function receives a timedelta to round to and gets the amount of seconds before that timestamp.

        :param timedelta timedelta_to_round: The timedelta to round to.
        :return: The seconds until the nearest rounded timedelta.
        :rtype: float
        """

        now = datetime.now()
        nearest = now + (datetime.min - now) % timedelta_to_round
        return nearest.timestamp() - now.timestamp()

    async def __detect_birthdays(self) -> None:
        await self.bot.wait_until_ready()

        while not self.bot.is_closed():
            await asyncio.sleep(self.round_to_nearest(timedelta(minutes=30)))

            for birthday_member in await self.get_members_with_birthday(
                self.get_midnight_timezones()
            ):
                guild = self.bot.get_guild(birthday_member["guild"])

                if guild:
                    member = guild.get_member(birthday_member["member"])

                    if member:
                        await self.call_event(
                            "on_member_birthday", BirthdayMember(self, member)
                        )



from __future__ import annotations

import asyncio
import logging
import os
import time
from typing import Optional, TYPE_CHECKING, List, Tuple

from disnake.ext import commands

if TYPE_CHECKING:
    from .base import DatabaseChecker
    from .database import Database

__all__ = ("ExtendedClient", "DatabaseClient", "ManagerClient")


class ExtendedClient(commands.Bot):
    """
    Represents an extended commands,Bot client.
    Adds a token attribute, replaces methods, loads cogs, etc.
    """

    __slots__ = ("token", "start_time")

    def __init__(self, token: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.token = token
        self.start_time = time.time()

    def load_cogs(self, directory: str, ignore_prefix: str = "__") -> None:
        """
        Loads all the cog extensions in the directory.

        :param str ignore_prefix: The prefix to ignore, files that start with that prefix will not be loaded.
        :param str directory: The directory.
        :return: None
        :rtype: None
        """

        extension_directory = directory.replace("/", ".")
        if extension_directory:
            extension_directory += "."

        working_directory = os.getcwd()

        slash = "/" if "/" in working_directory else "\\"

        for file in os.listdir(working_directory + f"{slash}{directory}"):
            if not file.endswith(".py") or file.startswith(ignore_prefix):
                continue

            try:
                self.load_extension(f"{extension_directory}{file.replace('.py', '')}")
                logging.info(f"Loaded cog {file}")
            except Exception as e:
                logging.error(f"Failed to load cog {file}")
                raise e

    def run(self, cogs_directory: Optional[str] = "cogs") -> None:
        """
        Runs the bot and loads the cogs automatically.

        :param Optional[str] cogs_directory: The directory to load the cogs from.
        :return: None
        :rtype: None
        """

        if cogs_directory is not None:  # Might be an empty string.
            self.load_cogs(cogs_directory)

        super().run(self.token)


class DatabaseClient(ExtendedClient):
    """
    Represents an extended client that has a database.
    """

    __slots__ = ("database",)

    def __init__(self, token: str, *args, **kwargs):
        super().__init__(token, *args, **kwargs)
        self.database: Optional[Database] = None

    async def wait_until_database_connection(self) -> None:
        """
        Waits until the database is connected.

        :return: None
        :rtype: None
        """

        while not self.database:
            await asyncio.sleep(0.1)


class ManagerClient(DatabaseClient):
    """
    Represents an extended database client that has managers.
    """

    __slots__ = ("managers",)

    def __init__(self, token: str, *args, **kwargs):
        super().__init__(token, *args, **kwargs)
        self.managers: List[Tuple[DatabaseChecker, Optional[List[str]]]] = []

        self.add_listener(self.on_ready)  # Doesnt do this automatically.

    def add_manager(self, manager: DatabaseChecker, tables: List[str] = None) -> None:
        self.managers.append((manager, tables))

    async def on_ready(self):
        await self.wait_until_database_connection()

        for manager, tables in self.managers:
            await manager.connect_to_database(self.database, tables or [])



import re

import youtube_dl

FFMPEG_OPTIONS = {
    "before_options": "-reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 5",
    "options": "-vn",
}

SPOTIFY_RE = re.compile("^https://open.spotify.com/")
DEEZER_RE = re.compile("^https://deezer.page.link/")
SOUNDCLOUD_RE = re.compile("^https://soundcloud.com/")

YTDL_OPTS = {
    "format": "bestaudio/best",
    "restrictfilenames": True,
    "noplaylist": False,
    "nocheckcertificate": True,
    "ignoreerrors": False,
    "logtostderr": False,
    "quiet": True,
    "no_warnings": True,
    "default_search": "auto",
}
YTDL = youtube_dl.YoutubeDL(YTDL_OPTS)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .playlist import Playlist
from .constants import *

if TYPE_CHECKING:
    from ..spotify import SpotifyClient
    from ..youtube import YoutubeClient


async def get_playlist(
    spotify: SpotifyClient, youtube: YoutubeClient, url: str
) -> Optional[Playlist]:
    if SPOTIFY_RE.match(url) and spotify:
        spotify_info = await spotify.fetch_full_playlist(url)
        return Playlist.from_spotify_dict(spotify_info) if spotify_info else None

    playlist_info = await youtube.get_playlist_information(
        await youtube.get_query_id(url)
    )
    return Playlist.from_youtube_dict(playlist_info) if playlist_info else None



from __future__ import annotations

import asyncio
from datetime import datetime
from typing import TYPE_CHECKING, List, Iterable

import aiohttp

from .base import DatabaseChecker

if TYPE_CHECKING:
    import disnake
    from disnake.ext import commands

__all__ = ("TwitchManager", "TwitchAuthorizationError", "get_twitch_oauth_key")


async def get_twitch_oauth_key(client_id: str, client_secret: str) -> str:
    """
    Gets the Twitch OAuth key from the Twitch API.

    :param client_id: The client ID.
    :param client_secret: The client secret.
    :return: The OAuth key.
    """

    async with aiohttp.ClientSession() as session:
        async with session.post(
            "https://id.twitch.tv/oauth2/token",
            data={
                "client_id": client_id,
                "client_secret": client_secret,
                "grant_type": "client_credentials",
            },
        ) as resp:
            resp_json = await resp.json()
            return resp_json["access_token"]


class TwitchAuthorizationError(Exception):
    """
    Raises an error when the Twitch API returns an authorization error.
    """


class TwitchManager(DatabaseChecker):
    """
    Represents a twitch notification manager.
    """

    __slots__ = (
        "bot",
        "update_interval",
        "twitch_client_id",
        "twitch_access_token",
        "session",
        "authorization_headers",
    )

    TWITCH_API_URL = "https://api.twitch.tv/helix"

    def __init__(
        self,
        bot: commands.Bot,
        twitch_client_id: str,
        twitch_access_token: str,
        update_interval: int = 30,
    ) -> None:
        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "channel": "string",
                }
            ],
            [
                "channels",
            ],
        )
        self.bot = bot

        self.twitch_client_id = twitch_client_id
        self.twitch_access_token = twitch_access_token

        self.authorization_headers = {
            "Client-ID": self.twitch_client_id,
            "Authorization": f"Bearer {self.twitch_access_token}",
        }

        self.update_interval = update_interval

        self._channel_cache: List[dict] = []
        self.session = None

        self.add_event(self._on_database_connect, "on_database_connect")

    async def _initialize(self) -> None:
        if not self.session:
            self.session = aiohttp.ClientSession()

    async def _on_database_connect(self):
        self.bot.loop.create_task(self.__detect_streams())

    async def get_channel_status(self, channels: Iterable[str]) -> list:
        """
        |coro|

        Gets the live status of the given channels.

        :param Iterable[str] channels: The channels.
        :return: The live statuses of the channels or the error dictionary.
        :rtype: list
        :raises: TwitchAuthorizationError: If the Twitch API returns an authorization error.
        """

        await self._initialize()

        url = (
            self.TWITCH_API_URL + f"/streams?user_login={'&user_login='.join(channels)}"
        )

        r = await self.session.get(
            url,
            headers=self.authorization_headers,
        )

        r_json = await r.json()
        if "data" in r_json:
            for stream in r_json["data"]:
                stream["started_at"] = datetime.fromisoformat(stream["started_at"][:-1])

            return r_json["data"]

        raise TwitchAuthorizationError(r_json["message"])

    async def add_channel(self, guild: disnake.Guild, channel: str) -> None:
        """
        |coro|

        Adds a channel to the guild list.

        :param disnake.Guild guild: The guild.
        :param str channel: The channel.
        :return: None
        :rtype: None
        """

        await self.database.insertifnotexists(
            self.tables["channels"],
            {"guild": guild.id, "channel": channel},
            {"guild": guild.id, "channel": channel},
        )

    async def remove_channel(self, guild: disnake.Guild, channel: str) -> None:
        """
        |coro|

        Removes a channel from the guild list.

        :param disnake.Guild guild: The guild.
        :param str channel: The channel.
        :return: None
        :rtype: None
        """

        await self.database.delete(
            self.tables["channels"], {"guild": guild.id, "channel": channel}
        )

    async def get_guild_channels(self, guild: disnake.Guild) -> List[str]:
        """
        |coro|

        Returns the channels of the guild.

        :param disnake.Guild guild: The guild.
        :return: The channels.
        :rtype: List[str]
        """

        channel_records = await self.database.select(
            self.tables["channels"], ["channel"], {"guild": guild.id}, True
        )

        return [record["channel"] for record in channel_records]

    @staticmethod
    def get_matching_channels(streams: List[dict], names: List[str]) -> List[dict]:
        """
        Returns the streams that match the given names.

        :param List[dict] streams: The streams.
        :param List[str] names: The names.
        :return: The streams that match the given names.
        :rtype: List[dict]
        """

        return [
            stream
            for stream in streams
            if stream["user_name"].casefold() in [x.casefold() for x in names]
        ]

    @staticmethod
    def remove_channel_ids(streams: List[dict], channel_ids: List[int]) -> List[dict]:
        """
        Removes the channels that are in the given list that match the channel ids.

        :param List[dict] streams: The streams.
        :param List[int] channel_ids: The channel ids.
        :return: The result channels.
        :rtype: List[dict]
        """

        return [stream for stream in streams if stream["id"] not in channel_ids]

    async def __detect_streams(self) -> None:
        await self.bot.wait_until_ready()

        start_time = datetime.utcnow()

        while not self.bot.is_closed():
            guild_channels = {
                x: await self.get_guild_channels(x) for x in self.bot.guilds
            }
            twitch_channels = {
                channel for channels in guild_channels.values() for channel in channels
            }

            statuses = await self.get_channel_status(twitch_channels)

            started_streams = self.remove_channel_ids(
                [status for status in statuses if start_time <= status["started_at"]],
                [x["id"] for x in self._channel_cache],
            )
            ended_streams = self.remove_channel_ids(
                self._channel_cache, [x["id"] for x in statuses]
            )

            for guild, channel_list in guild_channels.items():
                guild_started_streams = self.get_matching_channels(
                    started_streams, channel_list
                )
                guild_ended_streams = self.get_matching_channels(
                    ended_streams, channel_list
                )

                if guild_started_streams:
                    await self.call_event("on_stream", guild, guild_started_streams)
                if guild_ended_streams:
                    await self.call_event("on_stream_end", guild, guild_ended_streams)

            self._channel_cache = statuses

            await asyncio.sleep(self.update_interval)



from __future__ import annotations

import asyncio
from typing import Optional, TYPE_CHECKING, List, Iterable

if TYPE_CHECKING:
    from ..youtube import YoutubeClient
    import disnake


__all__ = ("Player",)


class Player:
    """
    Represents a music player.
    """

    __slots__ = (
        "data",
        "title",
        "stream_url",
        "url",
        "start_timestamp",
        "last_pause_timestamp",
        "duration",
        "requester",
        "source",
        "autoplayed",
        "used_query",
        "youtube_dl",
    )

    def __init__(
        self,
        requester: Optional[disnake.Member],
        used_query: str,
        title: str,
        stream_url: str,
        url: str,
        duration: int,
        data: dict,
        youtube_dl: bool = False,
    ):
        self.source = None
        self.data = data
        self.requester = requester
        self.title = title
        self.stream_url = stream_url
        self.url = url
        self.used_query = used_query

        self.autoplayed = False
        self.start_timestamp = 0
        self.last_pause_timestamp = 0

        self.youtube_dl = youtube_dl
        self.duration = duration if duration != 0 else "LIVE"

    def __str__(self):
        return self.title

    def __repr__(self):
        return f"<{self.__class__.__name__} requester={self.requester}, title={self.title}, duration={self.duration}>"

    @staticmethod
    def _get_stream_url(player: dict) -> Optional[str]:
        """
        Returns the stream url of a player.

        :param dict player: The player.
        :return: The stream url.
        :rtype: Optional[str]
        """

        if "adaptiveFormats" not in player["streamingData"]:
            return None

        stream_urls = [
            x
            for x in sorted(
                player["streamingData"]["adaptiveFormats"],
                key=lambda x: x.get("averageBitrate", 0),
                reverse=True,
            )
            if "audio" in x["mimeType"] and "opus" not in x["mimeType"]
        ]

        return player["streamingData"].get("hlsManifestUrl") or stream_urls[0]["url"]

    @classmethod
    async def make_multiple_players(
        cls,
        youtube: YoutubeClient,
        used_query: str,
        songs: Iterable[str],
        requester: Optional[disnake.Member],
    ) -> List[Player]:
        """
        |coro|

        Returns a list of players from a iterable of queries.

        :param str used_query: The used query.
        :param YoutubeClient youtube: The youtube client.
        :param requester: The requester.
        :type requester: Optional[disnake.Member]
        :param songs: The queries.
        :type songs: Iterable[str]
        :return: The list of created players.
        :rtype: List[Player]
        """

        tasks = [cls.fetch_song(youtube, song, playlist=False) for song in songs]

        songs = await asyncio.gather(*tasks)

        return [cls.create_player(requester, used_query, x) for x in songs if x]

    @classmethod
    async def get_similar_videos(
        cls, video_id: str, youtube: YoutubeClient
    ) -> List[Player]:
        """
        |coro|

        Creates similar videos related to the video id.

        :param str video_id: The video id
        :param YoutubeClient youtube: The youtube client.
        :return: The list of similar players
        :rtype: List[Player]
        """

        similar_video = await youtube.get_similar_videos(video_id)
        players = await cls.make_players(
            youtube, f"https://youtube.com/watch/?v={similar_video[0]}", None, False
        )
        for player in players:
            player.autoplayed = True

        return players

    @staticmethod
    async def fetch_data(
        youtube: YoutubeClient, query: str, playlist: bool = True
    ) -> List[dict]:
        """
        |coro|

        Fetches the youtube data of the query.

        :param YoutubeClient youtube: The youtube client.
        :param bool playlist: Indicating if it should fetch playlists.
        :param str query: The query.
        :return: The youtube data.
        :rtype: Optional[dict]
        """

        return [
            x
            for x in await youtube.get_videos(
                await youtube.get_query_id(query), playlist
            )
            if "streamingData" in x
        ]

    @classmethod
    async def fetch_song(
        cls, youtube: YoutubeClient, query: str, playlist: bool = True
    ) -> List[dict]:
        """
        |coro|

        Fetches the song's or playlist's data.
        Will return the first song in the playlist if playlist is False.

        :param YoutubeClient youtube: The youtube client.
        :param query: The query.
        :type query: str
        :param playlist: A bool indicating if the function should fetch playlists or get the first video.
        :type playlist: bool
        :return: The list of songs.
        :rtype: List[dict]
        """

        data = await cls.fetch_data(youtube, query, playlist)
        if not data:
            return []

        players = []
        for player in data:
            player["url"] = url = cls._get_stream_url(player)

            if url:
                players.append(
                    {x: y for x, y in player.items() if x in ["url", "videoDetails"]}
                )

        return players

    @classmethod
    def create_player(
        cls, requester: disnake.Member, query: str, player: dict
    ) -> Player:
        try:
            return cls(
                requester,
                query,
                player["videoDetails"]["title"],
                player.get("url"),
                "https://youtube.com/watch/?v=" + player["videoDetails"]["videoId"],
                int(player["videoDetails"]["lengthSeconds"]),
                data=player,
            )
        except TypeError:
            return cls(
                requester,
                query,
                player[0]["videoDetails"]["title"],
                player[0].get("url"),
                "https://youtube.com/watch/?v=" + player[0]["videoDetails"]["videoId"],
                int(player[0]["videoDetails"]["lengthSeconds"]),
                data=player,
            )


    @classmethod
    async def make_players(
        cls,
        youtube: YoutubeClient,
        query: str,
        requester: Optional[disnake.Member],
        playlist: bool = True,
    ) -> List[Player]:
        """
        |coro|

        Returns a list of players from the query.

        :param YoutubeClient youtube: The youtube client.
        :param Optional[disnake.Member] requester: The song requester.
        :param query: The query.
        :type query: str
        :param playlist: A bool indicating if the function should fetch playlists or get the first video.
        :type playlist: bool
        :return: The list of created players.
        :rtype: List[Player]
        """

        return [
            cls.create_player(requester, query, player)
            for player in await cls.fetch_song(youtube, query, playlist)
        ]



from __future__ import annotations

import disnake
from disnake.ext import commands
from typing import Union, List, Optional

from .base import DatabaseChecker


class ModMailManager(DatabaseChecker):
    def __init__(self, bot: Union[commands.Bot, commands.AutoShardedBot], trigger: str):
        self.bot = bot
        self.trigger = trigger

        super().__init__([{"guild": "snowflake", "channel": "snowflake"}], ["modmail"])

        self.bot.add_listener(self._handle_modmail_requests, "on_message")

    async def _handle_modmail_requests(self, message: disnake.Message):
        if not isinstance(message.channel, disnake.DMChannel):
            return

        if message.author.id == self.bot.user.id:
            return

        if self.trigger.lower() == (message.content.split())[0].lower():
            await self.call_event(
                "on_modmail_request",
                await self.bot.get_context(message=message, cls=commands.Context),
            )

    async def set_channel(self, channel: disnake.TextChannel) -> None:
        """
        :param channel: Channel that ModMail is sent to.
        :type channel: disnake.TextChannel
        :return:
        :rtype: None
        """

        self._check_database()

        table_data = {"guild": channel.guild.id, "channel": channel.id}

        await self.database.updateorinsert(
            self.tables["modmail"], table_data, {"guild": channel.guild.id}, table_data
        )

    async def get_channel(self, guild: disnake.Guild) -> disnake.TextChannel:
        """
        :param guild: The guild to fetch the ModMail Channel object for
        :type guild: disnake.Guild
        :return:
        :rtype: disnake.TextChannel
        """

        channel_id = await self.database.select(
            self.tables["modmail"], ["channel"], {"guild": guild.id}, fetchall=False
        )
        return self.bot.get_channel(channel_id["channel"])

    async def get_mutual_guilds(self, user: disnake.User) -> List[disnake.Guild]:
        """
        :param user: User to fetch the mutual guilds with the bot.
        :type user: disnake.User
        :return:
        :rtype: List[disnake.Guild]
        """

        return [x for x in self.bot.guilds if disnake.utils.get(x.members, id=user.id)]

    async def get_modmail_guild(
        self, ctx: commands.Context, guilds: List[disnake.Guild]
    ) -> Optional[disnake.Guild]:
        """
        :param ctx: Used to fetch channel
        :type ctx: commands.Context
        :param guilds: List of all mutual guilds
        :type guilds: List[disnake.Guild]
        :return:
        :rtype: disnake.Guild
        """
        embed = disnake.Embed(
            title="ModMail",
            description="Please type the Guild ID to send modmail to that server",
        )
        for guild in guilds:
            embed.add_field(name=f"{guild}", value=f"{guild.id}")

        def check(message):
            if isinstance(message.channel, disnake.DMChannel):
                return message.author.id == ctx.author.id

        await ctx.send(embed=embed)
        msg = await self.bot.wait_for("message", check=check, timeout=60)

        guildids = [guild.id for guild in guilds]

        if int(msg.content) in guildids:
            return self.bot.get_guild(int(msg.content))
        return None

    async def get_message(self, ctx: commands.Context) -> str:
        """
        :param ctx: fetch channel
        :type ctx: commands.Context
        :return:
        :rtype: str
        """

        embed = disnake.Embed(
            title="ModMail", description="Please type your message to the Mods."
        )
        await ctx.send(embed=embed)

        def check(message):
            if isinstance(message.channel, disnake.DMChannel):
                return message.author.id == ctx.author.id

        msg = await self.bot.wait_for("message", check=check, timeout=60)
        return msg.content



from __future__ import annotations

from typing import List, TYPE_CHECKING, Union, Any

from .enums import Loops

if TYPE_CHECKING:
    from ..youtube import YoutubeClient
    import disnake
    from .player import Player


__all__ = ("QueueManager",)


class QueueManager:
    """
    Represents a queue manager that manages a queue.
    """

    __slots__ = (
        "queue",
        "volume",
        "pos",
        "loop",
        "autoplay",
        "shuffle",
        "vote_skips",
        "played_history",
        "queue_loop_start",
        "original_queue",
    )

    def __init__(self, volume: float, queue: List[Player]):
        self.pos = -1
        self.queue = queue
        self.volume = volume
        self.autoplay = False
        self.shuffle = False
        self.queue_loop_start = 0
        self.loop = Loops.NO_LOOP
        self.vote_skips = []
        self.played_history: List[Player] = []
        self.original_queue: List[Player] = []

    async def get_next_player(self, youtube: YoutubeClient) -> Player:
        """
        |coro|

        Returns the next player that should be played from the queue.

        :param YoutubeClient youtube: The youtube client.
        :return: The player.
        :rtype: Player
        """

        if self.loop != Loops.LOOP:
            self.pos += 1

        if self.loop == Loops.LOOP:
            player = self.now_playing

        elif self.loop == Loops.QUEUE_LOOP:
            if self.is_finished():
                self.pos = self.queue_loop_start

            player = self.queue[self.pos]

        else:
            if not self.queue and self.autoplay:
                last_video_id = self.played_history[-1].data["videoDetails"]["videoId"]
                player = (await Player.get_similar_videos(last_video_id, youtube))[0]

            else:
                player = self.queue[self.pos]

        return player

    def is_finished(self) -> bool:
        """
        Returns a boolean representing if the queue is finished.

        :return: A boolean representing if the queue is finished.
        :rtype: bool
        """

        return self.pos >= len(self.queue)

    @property
    def player_queue(self) -> List[Player]:
        """
        Returns the player queue.

        :return: The list of players
        :rtype: List[Player]
        """

        return self.queue[self.pos + 1 :]

    @property
    def now_playing(self) -> Player:
        """
        Returns the currently playing song.

        :return: The currently playing song.
        :rtype: Player
        """

        return self.queue[self.pos]

    @property
    def history(self) -> List[Player]:
        """
        Returns the player history.

        :return: The history.
        :rtype: List[Player]
        """

        return self.queue[: self.pos]

    def add(self, player: Player) -> None:
        """
        Adds a player to the queue.

        :param player: The player to add.
        :type player: Player
        :return: None
        :rtype: None
        """

        self.queue.append(player)

    def clear(self) -> None:
        """
        Clears the queue.

        :return: None
        :rtype: None
        """

        self.queue = self.queue[: self.pos + 1]

    def remove(self, index: int) -> Union[Player, Any]:
        """
        Removes and element from the queue at the specified index, and returns the element's value.

        :param index: The index.
        :type index: int
        :return: The element's value
        :rtype: Union[Player, Any]
        """

        return self.queue.pop(index)

    def remove_member(self, member: disnake.Member) -> List[Player]:
        """
        Removes the member from the queue.

        :param disnake.Member member: The member to remove from the queue.
        :return: The players removed.
        :rtype: None
        """

        removed_players = []
        for player in self.player_queue:
            if player.requester == member:
                removed_players.append(player)
                self.queue.remove(player)

        return removed_players

    def cleanup(self):
        """
        Clears the queue.

        :return: None
        :rtype: None
        """

        self.clear()
        self.history.clear()
        del self.played_history
        del self.queue



from __future__ import annotations

from typing import Optional, TYPE_CHECKING, Dict

import wavelink

from .equalizer import Equalizer
from ..enums import ManagerType
from ..music import MusicManager

if TYPE_CHECKING:
    from disnake.ext import commands

__all__ = ("LavalinkMusicManager",)


class LavalinkMusicManager(MusicManager):
    """
    Represents a lavalink music manager.
    """

    def __init__(
        self,
        bot: commands.Bot,
        spotify_support: bool = True,
        inactivity_timeout: int = 60,
        minimum_users: int = 1,
        **kwargs,
    ):
        super().__init__(
            bot, spotify_support, inactivity_timeout, minimum_users, **kwargs
        )

        self.default_volume *= 100
        self.type = ManagerType.LAVALINK
        self.contexts: Dict[int, commands.Context] = {}
        self.bot.add_listener(self.__on_song_end, "on_wavelink_track_end")
        self.add_event(self.__on_queue_end, "on_queue_end")

    @staticmethod
    async def __on_queue_end(ctx):
        # MusicManager stops the voice client because the wavelink library assumes it is still playing.
        await ctx.voice_client.stop()

    async def __on_song_end(self, player: wavelink.Player, track, reason):
        await self._check_queue(self.contexts[player.guild.id])

    async def connect_node(
        self,
        host: str,
        password: str,
        port: int,
        identifier: str = "LavaLinkMusicManager",
    ) -> wavelink.Node:
        return await wavelink.NodePool.create_node(
            host=host, password=password, port=port, identifier=identifier, bot=self.bot
        )

    async def _check_queue(self, ctx: commands.Context) -> None:
        try:
            if not ctx.voice_client or not ctx.voice_client.is_connected():
                return

            queue = self.queue[ctx.guild.id]
            player = await queue.get_next_player(self.youtube)

            if player is None:
                await self.cleanup(None, ctx.guild)
                await self.call_event("on_queue_end", ctx)

            await ctx.voice_client.set_volume(queue.volume)
            await ctx.voice_client.play(
                (await wavelink.YouTubeTrack.search(player.url))[0]
            )
            self.contexts[ctx.guild.id] = ctx

            queue.played_history.append(player)
            queue.vote_skips = []
            await self.call_event("on_play", ctx, player)

        except (IndexError, KeyError):
            await self.cleanup(None, ctx.guild)
            await self.call_event("on_queue_end", ctx)

    @MusicManager.ensure_connection()
    async def get_player_played_duration(
        self, ctx: commands.Context, _=None
    ) -> Optional[float]:
        return ctx.voice_client.position

    @MusicManager.ensure_connection(check_playing=True, check_queue=True)
    async def volume(
        self, ctx: commands.Context, volume: int = None
    ) -> Optional[float]:
        if volume is None:
            return ctx.voice_client.volume

        await ctx.voice_client.set_volume(volume)
        self.queue[ctx.guild.id].volume = volume
        return ctx.voice_client.volume

    @MusicManager.ensure_connection(check_playing=True)
    async def get_equalizer(self, ctx: commands.Context) -> Optional[Equalizer]:
        """
        Returns the ctx's equalizer.

        :param commands.Context ctx: The context.
        :return: The equalizer.
        :rtype: Optional[Equalizer]
        """

        return ctx.voice_client.equalizer or Equalizer.flat()

    @MusicManager.ensure_connection(check_playing=True)
    async def set_equalizer(self, ctx: commands.Context, equalizer: Equalizer) -> bool:
        """
        |coro|

        Sets the ctx's equalizer.

        :param commands.Context ctx: The context.
        :param Equalizer equalizer: The equalizer.
        :return: A bool indicating if the set was successful,
        :rtype: Optional[bool]
        """

        await ctx.voice_client.set_eq(equalizer)
        return True

    @MusicManager.ensure_connection(check_playing=True)
    async def seek(self, ctx: commands.Context, position: int = 0) -> Optional[bool]:
        """
        |coro|

        Seeks the current player to the position (ms)

        :param ctx: The context
        :param position: time to seek to (ms)
        :return: A bool indicating if the seek was successful
        :rtype: Optional[bool]
        """

        await ctx.voice_client.seek(position=position)
        return True



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional

import aiohttp
import aiohttp.client_exceptions


__all__ = ("ServerNotFound", "FiveMPlayer", "FiveMServer")


class ServerNotFound(Exception):
    """Raises an error when a server is invalid or offline."""


@dataclass
class FiveMPlayer:
    """
    Represents a FiveM player.
    """

    id: int
    identifiers: Dict[str, str]
    name: str
    ping: int

    @classmethod
    def from_dict(cls, player_dict: dict) -> FiveMPlayer:
        """
        Creates a FiveM player object from a dict.

        :param dict player_dict: The player information.
        :return: The FiveM player.
        :rtype: FiveMPlayer
        """

        identifiers = dict([x.split(":") for x in player_dict["identifiers"]])
        return cls(
            player_dict["id"], identifiers, player_dict["name"], player_dict["ping"]
        )


@dataclass
class FiveMServer:
    """
    Represents a FiveM server.
    """

    ip: str
    resources: List[str]
    players: List[FiveMPlayer]
    name: str
    variables: Dict[str, str]

    @classmethod
    async def fetch(cls, ip: str) -> Optional[FiveMServer]:
        """
        |coro|

        Fetches the server and returns the server object.
        The server object includes players, resources, name, variables

        :param ip: The server IP.
        :return: The FiveM server.
        :rtype: Optional[FiveMServer]
        """

        base_address = "http://" + ip + "/"

        async with aiohttp.ClientSession() as session:
            try:
                await session.get(base_address)  # Server status check
            except (
                aiohttp.client_exceptions.ClientConnectorError,
                aiohttp.client_exceptions.InvalidURL,
            ):
                raise ServerNotFound(f"Server '{ip}' is invalid or offline.")

            players_request = await session.get(base_address + "players.json")
            info_request = await session.get(base_address + "info.json")
            dynamic_info_request = await session.get(base_address + "dynamic.json")

            info = await info_request.json(content_type=None)
            players = await players_request.json(content_type=None)
            dynamic = await dynamic_info_request.json(content_type=None)
            # This is still included in the session because parsing the json outside of it sometimes doesnt work
            # and block the program (?) :(

        return cls(
            ip,
            info["resources"],
            [FiveMPlayer.from_dict(player) for player in players],
            dynamic["hostname"],
            info["vars"],
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Dict, Any

__all__ = ("Equalizer",)


@dataclass
class Equalizer:
    """
    Represents a Equalizer that supports different voice effects.
    """

    raw: List[float]
    name: str

    @property
    def eq(self) -> List[Dict[str, Any]]:
        return [{"band": index, "gain": gain} for index, gain in enumerate(self.raw)]

    @classmethod
    def flat(cls) -> Equalizer:
        levels = [0.0 for i in range(15)]

        return cls(levels, "Flat")

    @classmethod
    def boost(cls):
        levels = [
            0.08,
            0.12,
            0.2,
            0.18,
            0.15,
            0.1,
            0.05,
            0.0,
            0.02,
            -0.04,
            -0.06,
            -0.08,
            -0.10,
            -0.12,
            -0.14,
        ]

        return cls(levels, "Boost")

    @classmethod
    def metal(cls):
        levels = [
            0.0,
            0.1,
            0.1,
            0.15,
            0.13,
            0.1,
            0.0,
            0.125,
            0.175,
            0.175,
            0.125,
            0.125,
            0.1,
            0.075,
            0.0,
        ]

        return cls(levels, "Metal")

    @classmethod
    def piano(cls):
        levels = [
            -0.25,
            -0.25,
            -0.125,
            0.0,
            0.25,
            0.25,
            0.0,
            -0.25,
            -0.25,
            0.0,
            0.0,
            0.5,
            0.25,
            -0.025,
        ]

        return cls(levels, "Piano")



from dataclasses import dataclass

import wavelink
import disnake

from .equalizer import Equalizer


@dataclass(init=False)
class LavalinkPlayer(wavelink.Player, disnake.VoiceProtocol):
    """
    Represents a LavalinkPlayer.
    """

    equalizer: Equalizer = Equalizer.flat()

    async def set_eq(self, equalizer: Equalizer) -> None:
        await self.node._websocket.send(
            op="equalizer", guildId=str(self.guild.id), bands=equalizer.eq
        )
        self.equalizer = equalizer



from __future__ import annotations

from dataclasses import dataclass

import disnake
from typing import List, Optional

from .base import DatabaseChecker


@dataclass
class EconomyAccount:
    """
    Represents an EconomyAccount.
    """

    economy_manager: EconomyManager
    member: disnake.Member

    def __post_init__(self):
        self.table = self.economy_manager.tables["economy"]

    @property
    def __checks(self):
        return EconomyManager.generate_checks(self.member)

    async def currency(self):
        currency_data = await self.economy_manager.database.select(
            self.table, ["currency"], self.__checks
        )
        return currency_data["currency"]

    async def bank(self):
        bank_data = await self.economy_manager.database.select(
            self.table, ["bank"], self.__checks
        )
        return bank_data["bank"]

    async def net(self):
        return await self.bank() + await self.currency()

    async def change_currency(self, amount: int):
        currency = await self.currency()
        await self.economy_manager.database.update(
            self.table, {"currency": currency + amount}, self.__checks
        )

    async def change_bank(self, amount: int):
        bank_amount = await self.bank()
        await self.economy_manager.database.update(
            self.table, {"bank": bank_amount + amount}, self.__checks
        )


class EconomyManager(DatabaseChecker):
    def __init__(self, bot):
        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "member": "snowflake",
                    "currency": "snowflake",
                    "bank": "snowflake",
                }
            ],
            ["economy"],
        )
        self.bot = bot

    @staticmethod
    def generate_checks(member: disnake.Member):
        return {"guild": member.guild.id, "member": member.id}

    async def create_account(self, member: disnake.Member) -> None:
        self._check_database()

        await self.database.insertifnotexists(
            self.tables["economy"],
            {"guild": member.guild.id, "member": member.id, "currency": 0, "bank": 0},
            self.generate_checks(member),
        )

    async def get_account(self, member: disnake.Member) -> Optional[EconomyAccount]:
        self._check_database()

        member_data = await self.database.select(
            self.tables["economy"],
            [],
            self.generate_checks(member),
            True,
        )

        if member_data:
            return EconomyAccount(self, member)

        return None

    @DatabaseChecker.uses_database
    async def get_leaderboard(self, guild: disnake.Guild) -> List[EconomyAccount]:
        guild_info = sorted(
            await self.database.select(
                self.tables["economy"], [], {"guild": guild.id}, True
            ),
            key=lambda x: x["bank"] + x["currency"],
            reverse=True,
        )

        members = []
        for member_info in guild_info:
            member = guild.get_member(member_info["member"])
            if member:
                members.append(EconomyAccount(self, member))

        return members



from __future__ import annotations

import datetime
import os
import textwrap
import time
from enum import Enum
from io import BytesIO
from typing import Optional, Tuple, Union, TYPE_CHECKING

import PIL
import PIL.ImageShow
import aiohttp
import disnake
from PIL import Image, ImageDraw, ImageFont
from PIL.ImageFont import FreeTypeFont

if TYPE_CHECKING:
    from .leveling import LevelingAccount


__all__ = ("ImageManager", "Backgrounds")


class ImageManager:
    """
    An image manager that manages picture creation.
    """

    __slots__ = ()

    DEFAULT_COLOR = (127, 255, 0)

    @staticmethod
    def load_asset(name: str) -> str:
        """
        Returns the asset path of the asset.

        :param str name: The asset.
        :return: The asset path.
        :rtype: str
        """

        return os.path.join(os.path.dirname(__file__), "assets", name)

    @staticmethod
    async def make_request(url: str) -> Optional[bytes]:
        """
        Returns the bytes of the URL response, if applicable.

        :param str url: The url.
        :return: The response bytes.
        :rtype: Optional[bytes]
        """

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                return await response.read()

    @classmethod
    async def convert_image(cls, url: str) -> Image.Image:
        """
        Converts the image to a PIL image.

        :param str url: The URL.
        :return: The converted image.
        :rtype: Image.Image
        """

        return PIL.Image.open(BytesIO(await cls.make_request(url))).convert("RGBA")

    @staticmethod
    def human_format(num: int) -> str:
        """
        Converts the number to a human readable format.

        :param int num: The number.
        :return: The human readable format.
        :rtype: str
        """

        original_num = num

        num = float("{:.3g}".format(num))
        magnitude = 0
        matches = ["", "K", "M", "B", "T", "Qua", "Qui"]
        while abs(num) >= 1000:
            if magnitude >= 5:
                break

            magnitude += 1
            num /= 1000.0

        try:
            return "{}{}".format(
                "{:f}".format(num).rstrip("0").rstrip("."), matches[magnitude]
            )
        except IndexError:
            return str(original_num)

    @staticmethod
    def multiline_text(
        card: Image.Image,
        text: str,
        font: FreeTypeFont,
        text_color: Tuple[int, int, int],
        start_height: Union[int, float],
        width: int,
    ) -> None:
        """
        Draws multiline text on the card.

        :param Image.Image card: The card to draw on.
        :param str text: The text to write.
        :param FreeTypeFont font: The font.
        :param Tuple[int, int, int] text_color: The text color.
        :param Union[int, float] start_height:
            The start height of the text, the text will start there, and make its way downwards.
        :param int width: The width of the wrap.
        :return: None
        :rtype: None
        """

        draw = ImageDraw.Draw(card)
        image_width, image_height = card.size

        y_text = start_height
        lines = textwrap.wrap(text, width=width)

        for line in lines:
            line_width, line_height = font.getsize(line)

            draw.text(
                ((image_width - line_width) / 2, y_text),
                line,
                font=font,
                fill=text_color,
            )

            y_text += line_height

    async def draw_profile_picture(
        self,
        card: Image.Image,
        member: disnake.Member,
        location: Tuple[int, int],
        size: int = 180,
        outline_thickness: int = 5,
        status: bool = True,
        outline_color: Tuple[int, int, int] = (255, 255, 255),
    ) -> Image.Image:
        """
        |coro|

        Pastes the profile picture on the card.

        :param Image.Image card: The card.
        :param disnake.Member member: The member to get the profile picture from.
        :param Tuple[int, int] location: The center of the picture.
        :param int size: The size of the pasted profile picture.
        :param int outline_thickness: The outline thickness.
        :param bool status: A bool indicating if it should paste the member's status icon.
        :param Tuple[int, int, int] outline_color: The outline color.
        :return: The result image.
        :rtype: Image.Image
        """

        blank = Image.new("RGBA", card.size, (255, 255, 255, 0))

        location = tuple(
            round(x - size / 2) if i <= 1 else round(x + size / 2)
            for i, x in enumerate(location + location)
        )

        outline_dimensions = tuple(
            x - outline_thickness if i <= 1 else x + outline_thickness
            for i, x in enumerate(location)
        )

        size_dimensions = (size, size)
        status_dimensions = tuple(round(x / 4) for x in size_dimensions)

        mask = Image.new("RGBA", card.size, 0)
        ImageDraw.Draw(mask).ellipse(location, fill=(255, 25, 255, 255))

        avatar = (await self.convert_image(str(member.avatar_url))).resize(
            size_dimensions
        )
        profile_pic_holder = Image.new("RGBA", card.size, (255, 255, 255, 255))

        ImageDraw.Draw(card).ellipse(outline_dimensions, fill=outline_color)

        profile_pic_holder.paste(avatar, location)
        pre_card = Image.composite(profile_pic_holder, card, mask)
        pre_card = pre_card.convert("RGBA")

        if status:
            status_picture = Image.open(self.load_asset(f"{member.status.name}.png"))
            status_picture = status_picture.convert("RGBA").resize(status_dimensions)

            blank.paste(
                status_picture, tuple(x - status_dimensions[0] for x in location[2:])
            )

        return Image.alpha_composite(pre_card, blank)

    async def create_welcome_card(
        self,
        member: disnake.Member,
        background: Union[Backgrounds, str],
        title: str,
        description: str,
        title_color: Tuple[int, int, int] = (255, 255, 255),
        description_color: Tuple[int, int, int] = (255, 255, 255),
        font_path: str = None,
        outline: int = 5,
        transparency: int = 0,
    ) -> disnake.File:
        """
        |coro|

        Creates a welcome image for the member and returns it as a disnake.File.

        :param disnake.Member member: The joined member.
        :param Union[Backgrounds, str] background: The background of the image, can be a Backgrounds enum or a URL.
        :param str title: The title.
        :param str description: The description.
        :param Tuple[int, int, int] title_color: The color of the title.
        :param Tuple[int, int, int] description_color: The color of the description.
        :param str font_path: The font path, uses the default font if not passed.
        :param int outline: The outline thickness.
        :param int transparency: The transparency of the background made.
        :return: The disnake file.
        :rtype: disnake.File
        """

        result_bytes = BytesIO()

        card = (
            Image.open(background.value)
            if isinstance(background, Backgrounds)
            else await self.convert_image(background)
        )
        card = card.resize((1024, 500))

        font_path = font_path if font_path else self.load_asset("font.ttf")

        big_font = ImageFont.truetype(font_path, 36)
        small_font = ImageFont.truetype(font_path, 30)

        draw = ImageDraw.Draw(card, "RGBA")
        if transparency and isinstance(background, Backgrounds):
            draw.rectangle((30, 30, 994, 470), fill=(0, 0, 0, transparency))

        draw.text((512, 360), title, title_color, font=big_font, anchor="ms")
        self.multiline_text(card, description, small_font, description_color, 380, 60)

        final_card = await self.draw_profile_picture(
            card, member, (512, 180), 260, outline_thickness=outline
        )

        final_card.save(result_bytes, format="PNG")
        result_bytes.seek(0)
        return disnake.File(result_bytes, filename="welcome_card.png")

    async def create_leveling_profile(
        self,
        member: disnake.Member,
        member_account: LevelingAccount,
        background: Union[Backgrounds, str],
        rank: int,
        name_color: Tuple[int, int, int] = DEFAULT_COLOR,
        rank_color: Tuple[int, int, int] = DEFAULT_COLOR,
        level_color: Tuple[int, int, int] = DEFAULT_COLOR,
        xp_color: Tuple[int, int, int] = DEFAULT_COLOR,
        bar_outline_color: Tuple[int, int, int] = (255, 255, 255),
        bar_fill_color: Tuple[int, int, int] = DEFAULT_COLOR,
        bar_blank_color: Tuple[int, int, int] = (255, 255, 255),
        profile_outline_color: Tuple[int, int, int] = DEFAULT_COLOR,
        font_path: str = None,
        outline: int = 5,
    ) -> disnake.File:
        """
        |coro|

        Creates a leveling image, converted to a disnake.File.

        :param disnake.Member member: The member.
        :param LevelingAccount member_account: The leveling account of the member.
        :param Union[Backgrounds, str] background: The background of the image.
        :param int rank: The guild rank of the member.
        :param Tuple[int, int, int] name_color: The color of the member's name.
        :param Tuple[int, int, int] rank_color: The color of the member's rank.
        :param Tuple[int, int, int] level_color: The color of the member's level.
        :param Tuple[int, int, int] xp_color: The color of the member's xp.
        :param Tuple[int, int, int] bar_outline_color: The color of the member's progress bar outline.
        :param Tuple[int, int, int] bar_fill_color: The color of the member's progress bar fill.
        :param Tuple[int, int, int] bar_blank_color: The color of the member's progress bar blank.
        :param Tuple[int, int, int] profile_outline_color: The color of the member's outliine.
        :param str font_path: The font path, uses the default font if not passed.
        :param int outline: The outline thickness.
        :return: The image, converted to a disnake.File.
        :rtype: disnake.File
        """

        result_bytes = BytesIO()

        card = (
            Image.open(background.value)
            if isinstance(background, Backgrounds)
            else await self.convert_image(background)
        )
        card = card.resize((850, 238))

        font_path = font_path if font_path else self.load_asset("font.ttf")
        font_big = ImageFont.truetype(font_path, 36)
        font_medium = ImageFont.truetype(font_path, 30)
        font_normal = ImageFont.truetype(font_path, 25)
        font_small = ImageFont.truetype(font_path, 20)

        draw = ImageDraw.Draw(card)
        draw.text((245, 90), str(member), name_color, font=font_big, anchor="ls")
        draw.text((800, 90), f"Rank #{rank}", rank_color, font=font_medium, anchor="rs")
        draw.text(
            (245, 165),
            f"Level {await member_account.level()}",
            level_color,
            font=font_normal,
            anchor="ls",
        )
        draw.text(
            (800, 165),
            f"{self.human_format(await member_account.xp())} /"
            f" {self.human_format(await member_account.next_level())} XP",
            xp_color,
            font=font_small,
            anchor="rs",
        )

        draw.rounded_rectangle(
            (242, 182, 803, 208),
            fill=bar_blank_color,
            outline=bar_outline_color,
            radius=13,
            width=3,
        )

        length_of_bar = await member_account.percentage_next_level() * 5.5 + 250
        draw.rounded_rectangle(
            (245, 185, length_of_bar, 205), fill=bar_fill_color, radius=10
        )

        final_card = await self.draw_profile_picture(
            card,
            member,
            (109, 119),
            outline_thickness=outline,
            outline_color=profile_outline_color,
        )

        final_card.save(result_bytes, format="PNG")
        result_bytes.seek(0)
        return disnake.File(result_bytes, filename="rankcard.png")

    async def create_spotify_card(
        self, spotify_activity: disnake.Spotify, font_path: str = None
    ) -> disnake.File:
        """
        |coro|

        Creates a Spotify activity image for the Spotify song and returns it as a disnake.File.

        :param disnake.Spotify spotify_activity: The Spotify activity.
        :param str font_path: The font path, uses the default font if not passed.
        :return: The disnake file.
        :rtype: disnake.File
        """

        result_bytes = BytesIO()

        album_image = await self.convert_image(spotify_activity.album_cover_url)

        # Background colour
        paletted = album_image.convert(
            "P", palette=Image.ADAPTIVE, colors=1
        )  # Reduce to palette
        # Finding dominant background color
        palette = paletted.getpalette()
        color_counts = paletted.getcolors()
        palette_index = color_counts[0][1]
        dominant_color = tuple(palette[palette_index * 3 : palette_index * 3 + 3])

        # Checking brightness of bg color
        brightness = (
            (0.21 * dominant_color[0])
            + (0.72 * dominant_color[1])
            + (0.07 * dominant_color[2])
        )

        if brightness < 100:
            text_color = "white"
            bg_img = "spotify_white.png"
        else:
            text_color = "black"
            bg_img = "spotify_black.png"

        track_background_image = Image.open(self.load_asset(bg_img))

        font_path = font_path if font_path else self.load_asset("font.ttf")

        # Fonts
        title_font = ImageFont.truetype(font_path, 16)
        artist_font = ImageFont.truetype(font_path, 14)
        album_font = ImageFont.truetype(font_path, 14)
        start_duration_font = ImageFont.truetype(font_path, 12)
        end_duration_font = ImageFont.truetype(font_path, 12)

        # Positions
        title_text_position = 150, 30
        artist_text_position = 150, 60
        album_text_position = 150, 80
        start_duration_text_position = 150, 119
        end_duration_text_position = 508, 119

        played_duration = (
            datetime.datetime.utcnow() - spotify_activity.start
        ).total_seconds()
        total_duration = spotify_activity.duration.total_seconds()

        played = played_duration / total_duration

        start_duration = time.strftime("%H:%M:%S", time.gmtime(played_duration))
        end_duration = time.strftime("%H:%M:%S", time.gmtime(total_duration))

        # Draws
        draw_on_image = ImageDraw.Draw(track_background_image)

        draw_on_image.text(
            title_text_position, spotify_activity.title, text_color, font=title_font
        )

        draw_on_image.text(
            artist_text_position,
            f"by {spotify_activity.artist}",
            text_color,
            font=artist_font,
        )

        draw_on_image.text(
            album_text_position, spotify_activity.album, text_color, font=album_font
        )

        draw_on_image.text(
            start_duration_text_position,
            start_duration,
            text_color,
            font=start_duration_font,
        )

        draw_on_image.text(
            end_duration_text_position, end_duration, text_color, font=end_duration_font
        )

        draw_on_image.rounded_rectangle(
            (198, 125, 198 + 300 * played, 129),
            fill=text_color,
            outline=None,
            radius=3,
            width=0,
        )

        background_image_color = Image.new(
            "RGBA", track_background_image.size, dominant_color
        )
        background_image_color.paste(
            track_background_image, (0, 0), track_background_image
        )

        # Resize
        album_image_resize = album_image.resize((140, 160))
        background_image_color.paste(album_image_resize, (0, 0), album_image_resize)

        # Save image
        background_image_color.convert("RGB")

        background_image_color.save(result_bytes, format="PNG")
        result_bytes.seek(0)
        return disnake.File(result_bytes, filename="spotify.png")


class Backgrounds(Enum):
    GALAXY = ImageManager.load_asset("1.png")
    BLANK_GRAY = ImageManager.load_asset("2.png")
    GAMING = ImageManager.load_asset("3.png")



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Any, TYPE_CHECKING, Union, Optional
from .enums import PlaylistType

if TYPE_CHECKING:
    import disnake
    from .music import MusicManager

__slots__ = ("SpotifyTrack", "YoutubeAuthor", "Playlist", "UserPlaylist")


@dataclass
class SpotifyTrack:
    """
    Represents a spotify track.
    """

    name: str
    authors: List[str]

    def __str__(self):
        return f"{self.name} by {self.authors[0]}"

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> SpotifyTrack:
        """
        Creates a Spotify track from the dictionary.

        :param Dict[str, Any] dictionary: The dictionary.
        :return: The spotify track.
        :rtype: SpotifyTrack
        """

        return cls(
            dictionary["track"]["name"],
            [artist["name"] for artist in dictionary["track"]["artists"]],
        )


@dataclass
class YoutubeAuthor:
    """
    Represents a YouTube author / channel.
    """

    name: str
    id: str

    @classmethod
    def from_dict(cls, dictionary: Dict[str, str]) -> YoutubeAuthor:
        """
        Creates a YouTube author from the dictionary.

        :param Dict[str, str] dictionary: The dictionary.
        :return: The YouTube author.
        :rtype: YoutubeAuthor
        """

        return cls(dictionary["name"], dictionary["id"])


@dataclass
class Playlist:
    """
    Represents a playlist.
    Supports Spotify and YouTube.
    """

    title: str
    author: Optional[YoutubeAuthor]
    songs: List[Union[str, SpotifyTrack]]
    url: str
    type: PlaylistType

    @classmethod
    def from_youtube_dict(cls, dictionary: Dict[str, Any]) -> Playlist:
        """
        Creates a playlist object from the YouTube dictionary

        :param Dict[str, Any] dictionary: The YouTube dictionary.
        :return: The playlist.
        :rtype: Playlist
        """

        return cls(
            dictionary["title"],
            YoutubeAuthor.from_dict(dictionary["channel"]),
            dictionary["songs"],
            f"https://www.youtube.com/watch?v={dictionary['songs'][0]}&list={dictionary['playlistId']}",
            PlaylistType.YOUTUBE,
        )

    @classmethod
    def from_spotify_dict(cls, dictionary: Dict[str, Any]) -> Playlist:
        """
        Creates a playlist object from the Spotify dictionary

        :param Dict[str, Any] dictionary: The spotify dictionary.
        :return: The playlist.
        :rtype: Playlist
        """

        return cls(
            dictionary["name"],
            None,
            [SpotifyTrack.from_dict(track) for track in dictionary["tracks"]],
            dictionary["url"],
            PlaylistType.SPOTIFY,
        )


@dataclass
class UserPlaylist:
    """
    Represents a playlist stored in the database.
    """

    music_manager: MusicManager
    owner: disnake.User
    id: str
    playlist: Playlist

    async def delete(self) -> None:
        """
        |coro|

        Deletes the playlist from the database.

        :return: None
        :rtype: None
        """

        await self.music_manager.database.delete(
            self.music_manager.tables["playlists"],
            {"user": self.owner.id, "id": self.id},
        )



from typing import Union, Any, Tuple, Iterable

import disnake
from disnake.ext import commands

from .base import DatabaseChecker


class PrefixManager(DatabaseChecker):
    """
    A prefix manager that manages prefixes.

    :ivar commands.Bot bot: The bot.
    :ivar Iterable[str] default_prefixes: The default prefixes.
    :ivar bool mentioned: A bool indicating if the manager should support commands.when_mentioned_or.
    :ivar dict prefix_cache: The prefix cache.
    """

    def __init__(
        self,
        bot: commands.Bot,
        default_prefixes: Iterable[str],
        mentioned: bool = False,
    ):
        super().__init__([{"guild": "snowflake", "prefix": "string"}], ["prefixes"])
        self.default_prefixes = default_prefixes
        self.bot = bot
        self.mentioned = mentioned

        self.prefix_cache = {}
        bot.command_prefix = self.__get_prefix

    @DatabaseChecker.uses_database
    async def get_prefix(self, guild: Union[disnake.Guild, Any]) -> Iterable[str]:
        """
        |coro|

        Returns the prefix or the guild or the default prefixes.

        :param disnake.Guild guild: The guild.
        :return: The prefixes.
        :rtype: Iterable[str]
        """

        if guild.id in self.prefix_cache:
            return self.prefix_cache[guild.id]

        prefix = await self.database.select(
            self.tables["prefixes"], ["prefix"], {"guild": guild.id}
        )
        prefix = (prefix["prefix"],) if prefix else self.default_prefixes

        self.prefix_cache[guild.id] = prefix

        return prefix

    @DatabaseChecker.uses_database
    async def delete_prefix(self, guild: disnake.Guild) -> None:
        """
        |coro|

        Deletes the prefix record of the guild.

        :param disnake.Guild guild: The guild.
        :return: None
        :rtype: None
        """

        self.prefix_cache.pop(guild.id)
        await self.database.delete(self.tables["prefixes"], {"guild": guild.id})

    @DatabaseChecker.uses_database
    async def set_prefix(self, guild: disnake.Guild, prefix: str) -> None:
        """
        |coro|

        Sets the prefix of the guild.

        :param disnake.Guild guild: The guild.
        :param str prefix: The prefix that should be set.
        :return: None
        :rtype: None
        """

        if prefix in self.default_prefixes:
            await self.delete_prefix(guild)
            return

        self.prefix_cache[guild.id] = (prefix,)
        await self.database.updateorinsert(
            self.tables["prefixes"],
            {"prefix": prefix},
            {"guild": guild.id},
            {"guild": guild.id, "prefix": prefix},
        )

    async def __get_prefix(self, bot, message: disnake.Message) -> Tuple[str]:
        if not message.guild:
            return (
                commands.when_mentioned_or(*self.default_prefixes)(bot, message)
                if self.mentioned
                else self.default_prefixes
            )

        prefix = await self.get_prefix(message.guild)

        return (
            commands.when_mentioned_or(*prefix)(bot, message)
            if self.mentioned
            else prefix
        )



from __future__ import annotations

import asyncio
import random
import time
import uuid
from typing import Optional, TYPE_CHECKING, List, Tuple, Dict, Callable, Union

import aiohttp
import disnake

from .constants import *
from .enums import Loops, ManagerType
from .exceptions import (
    QueueEmpty,
    NotPlaying,
    NotConnected,
    RemoveIndexInvalid,
    AlreadyPaused,
    NotPaused,
    InvalidSkipIndex,
    SkipError,
    AlreadyConnected,
    UserNotConnected,
    InvalidPreviousIndex,
)
from .lavalink.player import LavalinkPlayer
from .player import Player
from .playlist import Playlist, UserPlaylist
from .queue import QueueManager
from .utils import get_playlist
from ..base import create_task, DatabaseChecker, maybe_coroutine
from ..spotify import SpotifyClient
from ..youtube import YoutubeClient

if TYPE_CHECKING:
    from disnake.ext import commands

__all__ = ("MusicManager",)


class MusicManager(DatabaseChecker):
    """
    Represents a MusicManager.
    """

    __slots__ = (
        "bot",
        "client_id",
        "client_secret",
        "spotify_support",
        "inactivity_timeout",
        "queue",
        "spotify",
    )

    def __init__(
        self,
        bot: commands.Bot,
        spotify_support: bool = True,
        inactivity_timeout: int = 60,
        minimum_users: int = 1,
        opus_players: bool = False,
        **kwargs,
    ):
        super().__init__(
            [
                {
                    "user": "snowflake",
                    "playlist_url": "string",
                    "id": "string",
                }
            ],
            ["playlists"],
        )
        self.bot = bot
        self.bot.add_listener(self._on_voice_state_update, "on_voice_state_update")

        self.client_id = kwargs.get("client_id")
        self.client_secret = kwargs.get("client_secret")
        self.default_volume = kwargs.get("default_volume") or 0.1
        self.executable = kwargs.get("executable") or "ffmpeg"
        self.spotify_support = spotify_support
        self.inactivity_timeout = 0 if not inactivity_timeout else inactivity_timeout
        self.minimum_users = minimum_users

        self.queue: Dict[int, QueueManager] = {}
        self.youtube = YoutubeClient(loop=self.bot.loop)
        self.opus_players = opus_players

        self._load_opus()

        if spotify_support:
            self.spotify = SpotifyClient(
                client_id=self.client_id,
                client_secret=self.client_secret,
                loop=self.bot.loop,
            )

        self.type = ManagerType.FFMPEG

    @staticmethod
    def _load_opus() -> None:
        """
        Ensures the opus library is loaded.

        :return: None
        :rtype: None
        :raises: RuntimeError: Could not find opus on the machine.
        """

        if not disnake.opus.is_loaded():
            try:
                disnake.opus._load_default()
            except OSError:
                raise RuntimeError("Could not find opus on the machine.")

    async def cleanup(
        self, voice_client: Optional[disnake.VoiceClient], guild: disnake.Guild
    ):
        """
        |coro|

        Cleans up after a guild.

        :param disnake.Guild guild: The guild to cleanup.
        :param Optional[disnake.VoiceClient] voice_client: The voice client.
        :return: None
        :rtype: None
        """

        if voice_client:
            try:
                await voice_client.disconnect(force=True)
            except ValueError:
                # Raised from wavelink
                pass

        if guild.id in self.queue:
            queue = self.queue.pop(guild.id)
            queue.cleanup()
            del queue

    @DatabaseChecker.uses_database
    async def add_playlist(
        self, user: disnake.User, url: str
    ) -> Optional[UserPlaylist]:
        """
        |coro|

        Adds a playlist to the user's account.
        Saves the playlist in the database.

        :param disnake.User user: The owner of the playlist.
        :param str url: The playlist URL.
        :return: None
        :rtype: None
        """

        playlist = await get_playlist(self.spotify, self.youtube, url)

        if not playlist:
            return

        generated_id = str(uuid.uuid4())
        await self.database.insertifnotexists(
            self.tables["playlists"],
            {"user": user.id, "playlist_url": url, "id": generated_id},
            {"user": user.id, "playlist_url": url},
        )

        return UserPlaylist(self, user, generated_id, playlist)

    @DatabaseChecker.uses_database
    async def get_playlist(
        self, user: disnake.User, playlist_id: str, partial: bool = False
    ) -> Optional[UserPlaylist]:
        """
        |coro|

        Gets a user playlist by id.

        :param str playlist_id: The playlist id.
        :param bool partial: Indicating if the function should not fetch the playlist data.
        :param disnake.User user: The user.
        :return: The user playlist.
        :rtype: Optional[UserPlaylist]
        """

        playlist = await self.database.select(
            self.tables["playlists"], [], {"user": user.id, "id": playlist_id}
        )

        if playlist:
            return UserPlaylist(
                self,
                user,
                playlist_id,
                await get_playlist(self.spotify, self.youtube, playlist["playlist_url"])
                if not partial
                else None,
            )

    @DatabaseChecker.uses_database
    async def get_user_playlists(
        self, user: disnake.User, partial: bool = False
    ) -> List[UserPlaylist]:
        """
        |coro|

        Returns the user's playlists.

        :param disnake.User user: The user.
        :param bool partial: Indicating if the function should not fetch the playlist data.
        :return: The list of user playlists.
        :rtype: List[UserPlaylist]
        """

        user_playlist_ids = await self.database.select(
            self.tables["playlists"], ["id"], {"user": user.id}, True
        )

        return list(
            await asyncio.gather(
                *[
                    self.get_playlist(user, user_playlist_id["id"], partial)
                    for user_playlist_id in user_playlist_ids
                ]
            )
        )

    async def _on_voice_state_update(self, member, before, after):
        voice_client = member.guild.voice_client
        channel_change = before.channel != after.channel

        if member == self.bot.user and channel_change and before.channel:
            await self.cleanup(voice_client, member.guild)

        elif voice_client and channel_change:
            voice_members = list(
                filter(lambda x: not x.bot, voice_client.channel.members)
            )

            if len(voice_members) < self.minimum_users:
                await asyncio.sleep(self.inactivity_timeout)

                await self.cleanup(voice_client, member.guild)

    async def ensure_activity(self, ctx: commands.Context) -> None:
        """
        |coro|

        Waits the inactivity timeout and ensures the voice client in ctx is playing a song.
        If no song is playing, it disconnects and calls the on_inactivity_timeout event.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: None
        :rtype: None
        """

        if self.inactivity_timeout is None:
            return

        await asyncio.sleep(self.inactivity_timeout)

        if (
            ctx.voice_client
            and ctx.voice_client.is_connected()
            and not ctx.voice_client.is_playing()
        ):
            await self.cleanup(ctx.voice_client, ctx.guild)

            await self.call_event("on_inactivity_disconnect", ctx)

    async def _check_connection(
        self,
        ctx: commands.Context,
        check_playing: bool = False,
        check_queue: bool = False,
    ) -> Optional[bool]:
        """
        |coro|

        Checks the connection state of the voice client in ctx.

        :param ctx: The context.
        :type ctx: commands.Context
        :param check_playing: A bool indicating if the function should check if a song is playing.
        :type check_playing: bool
        :param check_queue: A bool indicating if the function should check if a queue exists.
        :type check_queue: bool
        :return: True if all the checks passed.
        :rtype: bool
        """

        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event(
                "on_music_error",
                ctx,
                NotConnected("Client is not connected to a voice channel"),
            )
            return

        if check_playing and not ctx.voice_client.is_playing():
            await self.call_event(
                "on_music_error",
                ctx,
                NotPlaying("Client is not playing anything currently"),
            )
            return

        if check_queue and ctx.guild.id not in self.queue:
            await self.call_event("on_music_error", ctx, QueueEmpty("Queue is empty"))
            return

        return True

    def ensure_connection(*d_args, **d_kwargs) -> Callable:
        """
        A decorator which ensures there is a proper connection before invoking the decorated function.

        :param d_args: The connection arguments.
        :param d_kwargs: The connection key arguments.
        :return: The decorator.
        :rtype: Callable
        """

        def decorator(function):
            async def wrapper(self, ctx, *args, **kwargs):
                if await self._check_connection(ctx, *d_args, **d_kwargs):
                    return await function(self, ctx, *args, **kwargs)

            return wrapper

        return decorator

    async def _check_queue(self, ctx: commands.Context) -> None:
        """
        |coro|

        Plays the next song in the queue, handles looping, queue looping, autoplay, etc.

        :param ctx: The context of the voice client.
        :type ctx: commands.Context
        :return: None
        :rtype: None
        """

        try:
            if not ctx.voice_client or not ctx.voice_client.is_connected():
                return

            queue = self.queue[ctx.guild.id]
            player = await queue.get_next_player(self.youtube)

            if player is None:
                await self.cleanup(None, ctx.guild)
                await self.call_event("on_queue_end", ctx)

            player.source = (
                disnake.PCMVolumeTransformer(
                    disnake.FFmpegPCMAudio(
                        player.stream_url, **FFMPEG_OPTIONS, executable=self.executable
                    ),
                    queue.volume,
                )
                if not self.opus_players
                else disnake.FFmpegOpusAudio(
                    player.stream_url, **FFMPEG_OPTIONS, executable=self.executable
                )
            )

            ctx.voice_client.play(
                player.source,
                after=lambda x: create_task(self.bot.loop, self._check_queue(ctx)),
            )

            player.start_timestamp = time.time()

            queue.played_history.append(player)
            queue.vote_skips = []
            await self.call_event("on_play", ctx, player)

        except (IndexError, KeyError):
            await self.cleanup(None, ctx.guild)
            await self.call_event("on_queue_end", ctx)

    async def get_player_playlist(self, player: Player) -> Optional[Playlist]:
        """
        |coro|

        Returns the player's playlist, if applicable.

        :param Player player: The player.
        :return: The player's playlist.
        :rtype: Optional[Playlist]
        """

        return await get_playlist(self.spotify, self.youtube, player.used_query)

    @ensure_connection()
    async def get_player_played_duration(
        self, ctx: commands.Context, player: Player
    ) -> Optional[float]:
        """
        |coro|

        Returns the played duration of a player.

        :param ctx: The context.
        :type ctx: commands.Context
        :param player: The player.
        :type player: Player
        :return: The played duration of the player in seconds.
        :rtype: Optional[float]
        """

        start_timestamp = player.start_timestamp
        if ctx.voice_client.is_paused():
            start_timestamp = (
                player.start_timestamp + time.time() - player.last_pause_timestamp
            )

        time_played = time.time() - start_timestamp
        return min(
            time_played, time_played if player.duration == "LIVE" else player.duration
        )

    async def create_playlist_players(
        self, playlist: Playlist, requester: disnake.Member
    ) -> List[Player]:
        """
        |coro|

        Returns a list of players from the playlist.

        :param Playlist playlist: The playlist.
        :param disnake.Member requester: The requester.
        :return: The list of created players.
        :rtype: List[Player]
        """

        return await Player.make_multiple_players(
            self.youtube,
            playlist.url,
            [
                str(song) for song in playlist.songs
            ],  # Converts the song to str to convert any spotify tracks.
            requester,
        )

    @staticmethod
    async def fetch_ytdl_data(url: str) -> Optional[dict]:
        """
        |coro|

        Fetches the data from a url.

        :param str url: The url to fetch the data from.
        :return: The data from the url if applicable.
        :rtype: Optional[dict]
        """

        try:
            loop = asyncio.get_event_loop()
            return await loop.run_in_executor(
                None, lambda: YTDL.extract_info(url, download=False)
            )
        except youtube_dl.utils.DownloadError:
            return None

    async def create_player(
        self, query: str, requester: disnake.Member
    ) -> List[Player]:
        """
        |coro|

        Creates a list of players from the query.
        This function supports Spotify and all YTDL supported links.

        :param requester: The requester.
        :type requester: disnake.Member
        :param query: The query.
        :type query: str
        :return: The list of players.
        :rtype: List[Player]
        """

        if SPOTIFY_RE.match(query) and self.spotify_support:
            return await Player.make_multiple_players(
                self.youtube,
                query,
                [song for song in await self.spotify.get_songs(query)],
                requester,
            )

        if DEEZER_RE.match(query) or SOUNDCLOUD_RE.match(query):
            data = await self.fetch_ytdl_data(query)
            if not data:
                return []

            return [
                Player(
                    requester,
                    query,
                    data["title"],
                    data["url"],
                    data["webpage_url"],
                    data.get("duration", 30),
                    data,
                    True,
                )
            ]

        return await Player.make_players(self.youtube, query, requester)

    @ensure_connection()
    async def queue_add(
        self,
        ctx: commands.Context,
        players: List[Player],
    ) -> Optional[bool]:
        """
        |coro|

        Adds a list of players to the ctx queue.
        If a queue does not exist in ctx, it creates one.

        :param players: The list of players.
        :type players: List[Player]
        :param ctx: The context.
        :type ctx: commands.Context
        :return: A bool indicating if it was successful
        :rtype: Optional[bool]
        """

        if ctx.guild.id in self.queue:
            self.queue[ctx.guild.id].queue += players
        else:
            self.queue[ctx.guild.id] = QueueManager(self.default_volume, players)

        return True

    @ensure_connection(check_queue=True)
    async def queue_remove(self, ctx: commands.Context, index: int) -> Optional[Player]:
        """
        |coro|

        Removes a player from the queue in ctx at the specified index.
        Calls on_music_error with RemoveIndexInvalid if index is invalid.

        :param ctx: The context.
        :type ctx: commands.Context
        :param index: The index.
        :type index: int
        :return: The player that was removed, if applicable.
        :rtype: Optional[Player]
        """

        try:
            queue = self.queue[ctx.guild.id]

            return queue.remove(queue.pos + index)
        except IndexError:
            await self.call_event(
                "on_music_error",
                ctx,
                RemoveIndexInvalid("Failure when removing player from queue"),
            )

    async def lyrics(
        self, ctx: commands.Context, query: str = None
    ) -> Optional[Tuple[str, str, str]]:
        """
        |coro|

        Returns the lyrics from the query or the currently playing song.

        :param ctx: The context.
        :type ctx: commands.Context
        :param query: The query.
        :type query: str
        :return: The lyrics and the song name.
        :rtype: Optional[Tuple[str, str, str]]
        """

        query = await self.now_playing(ctx) if query is None else query
        if not query:
            return

        url = f"https://some-random-api.ml/lyrics?title={query}"

        async with aiohttp.ClientSession() as session:
            request = await session.get(url)
            request_json = await request.json(content_type=None)

            authors = request_json.get("author")
            title = request_json.get("title")
            lyrics = request_json.get("lyrics")

            return (title, authors, lyrics) if lyrics else None

    @ensure_connection()
    async def play(
        self,
        ctx: commands.Context,
    ) -> Optional[bool]:
        """
        |coro|

        Plays the player or the next song in the queue.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: A bool indicating if the play was successful
        :rtype: Optional[bool]
        """

        if not ctx.voice_client.is_playing():
            await self._check_queue(ctx)
            return True

    @ensure_connection()
    async def pause(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Pauses the currently playing song in ctx.
        Calls on_music_error with AlreadyPaused if already paused.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: A bool indicating if the pause was successful
        :rtype: Optional[bool]
        """

        if ctx.voice_client.is_paused():
            await self.call_event(
                "on_music_error", ctx, AlreadyPaused("Player is already paused.")
            )
            return

        if self.type == ManagerType.LAVALINK:
            await ctx.voice_client.set_pause(pause=True)
        else:
            (await self.now_playing(ctx)).last_pause_timestamp = time.time()
            ctx.voice_client.pause()

        create_task(self.bot.loop, self.ensure_activity(ctx))
        return True

    @ensure_connection()
    async def resume(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Resumes the currently paused song in ctx.
        Calls on_music_error with NotPaused if not paused.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: A bool indicating if the resume was successful
        :rtype: Optional[bool]
        """

        if not ctx.voice_client.is_paused():
            await self.call_event(
                "on_music_error", ctx, NotPaused("Player is not paused")
            )
            return

        if self.type == ManagerType.LAVALINK:
            await ctx.voice_client.set_pause(pause=False)
        else:
            ctx.voice_client.resume()

            now_playing = await self.now_playing(ctx)
            now_playing.start_timestamp += (
                time.time() - now_playing.last_pause_timestamp
            )

        return True

    @ensure_connection(check_playing=True, check_queue=True)
    async def previous(
        self, ctx: commands.Context, index: int = None, no_autoplay: bool = False
    ) -> Optional[List[Player]]:
        """
        |coro|

        Plays the (index) player from the history.

        :param commands.Context ctx: The ctx.
        :param bool no_autoplay: A bool indicating if autoplayed songs should be added back to the queue.
        :param int index: The index.
        :return: The list of Players that have been added back.
        :rtype: Optional[List[Player]]
        """

        queue = self.queue[ctx.guild.id]

        previous_index = 2 if index is None else index + 1
        if 0 >= previous_index:
            if index:
                await self.call_event(
                    "on_music_error",
                    ctx,
                    InvalidPreviousIndex("Previous index invalid."),
                )
                return

        original_queue_position = queue.pos
        queue.pos -= previous_index
        previous_players = queue.queue[queue.pos + 1 : original_queue_position]

        if no_autoplay:
            for player in previous_players[:]:
                if not player.requester:
                    previous_players.remove(player)
                    queue.queue.remove(player)

        await maybe_coroutine(ctx.voice_client.stop)
        return previous_players

    @ensure_connection(check_playing=True, check_queue=True)
    async def goto(self, ctx: commands.Context, index: int = 0) -> None:
        queue = self.queue[ctx.guild.id]

        queue.pos = index
        await maybe_coroutine(ctx.voice_client.stop)

    @ensure_connection(check_playing=True, check_queue=True)
    async def skip(self, ctx: commands.Context, index: int = None) -> Optional[Player]:
        """
        |coro|

        Skips to the index in ctx.
        Calls on_music_error with InvalidSkipIndex or SkipError.

        :param index: The index to skip to.
        :type index: int
        :param ctx: The context.
        :type ctx: commands.Context
        :return: The skipped player if applicable.
        :rtype: Optional[Player]
        """

        queue = self.queue[ctx.guild.id]

        # Created duplicate to make sure InvalidSkipIndex isn't raised when the user does pass an index and the queue
        # is empty.
        skip_index = 0 if index is None else index - 1
        if not skip_index < len(queue.queue) and not queue.pos < skip_index:
            if index:
                await self.call_event(
                    "on_music_error", ctx, InvalidSkipIndex("Skip index invalid.")
                )
                return

        if (
            not queue.autoplay
            and queue.loop != Loops.QUEUE_LOOP
            and (len(queue.queue) - 1) <= queue.pos + skip_index
        ):
            await self.call_event(
                "on_music_error", ctx, SkipError("No song to skip to.")
            )
            return

        original_position = queue.pos
        queue.pos += skip_index

        if queue.autoplay:
            last_video_id = queue.played_history[-1].data["videoDetails"]["videoId"]
            player = (await Player.get_similar_videos(last_video_id, self.youtube))[0]
            queue.add(player)
        else:
            player = queue.queue[original_position]

        await maybe_coroutine(ctx.voice_client.stop)
        return player

    @ensure_connection(check_playing=True, check_queue=True)
    async def volume(
        self, ctx: commands.Context, volume: int = None
    ) -> Optional[float]:
        """
        |coro|

        Sets the volume in ctx.
        Returns the current volume if volume is None.

        :param volume: The volume to set.
        :type volume: int
        :param ctx: The context.
        :type ctx: commands.Context
        :return: The new volume.
        :rtype: Optional[float]
        """

        if volume is None:
            return ctx.voice_client.source.volume * 100

        ctx.voice_client.source.volume = volume / 100
        self.queue[ctx.guild.id].volume = volume / 100
        return ctx.voice_client.source.volume * 100

    async def join(self, ctx: commands.Context) -> Optional[disnake.VoiceChannel]:
        """
        |coro|

        Joins the ctx voice channel.
        Calls on_music_error with AlreadyConnected or UserNotConnected.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: The voice channel it joined.
        :rtype: Optional[disnake.VoiceChannel]
        """

        if ctx.voice_client and ctx.voice_client.is_connected():
            await self.call_event(
                "on_music_error",
                ctx,
                AlreadyConnected("Client is already connected to a voice channel"),
            )
            return

        if not ctx.author.voice:
            await self.call_event(
                "on_music_error",
                ctx,
                UserNotConnected("User is not connected to a voice channel"),
            )
            return

        channel = ctx.author.voice.channel
        await channel.connect(
            cls=LavalinkPlayer
            if self.type == ManagerType.LAVALINK
            else disnake.VoiceClient
        )
        return channel

    @ensure_connection()
    async def leave(self, ctx: commands.Context) -> Optional[disnake.VoiceChannel]:
        """
        |coro|

        Leaves the voice channel in ctx.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: The voice channel it left.
        :rtype: Optional[disnake.VoiceChannel]
        """

        if ctx.guild.id in self.queue:
            self.queue[ctx.guild.id].cleanup()
            del self.queue[ctx.guild.id]

        await maybe_coroutine(ctx.voice_client.stop)

        channel = ctx.voice_client.channel
        await ctx.voice_client.disconnect(force=True)
        return channel

    @ensure_connection()
    async def stop(self, ctx: commands.Context) -> Optional[disnake.VoiceChannel]:
        """
        |coro|

        Stop the player in ctx.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: The voice channel it left.
        :rtype: Optional[disnake.VoiceChannel]
        """

        if ctx.guild.id in self.queue:
            self.queue[ctx.guild.id].cleanup()
            del self.queue[ctx.guild.id]

        await maybe_coroutine(ctx.voice_client.stop)

        channel = ctx.voice_client.channel
        return channel

    @ensure_connection(check_queue=True)
    async def now_playing(self, ctx: commands.Context) -> Optional[Player]:
        """
        |coro|

        Returns the currently playing player.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: The currently playing player.
        :rtype: Optional[Player]
        """

        now_playing = self.queue[ctx.guild.id].now_playing
        if not ctx.voice_client.is_playing() and not ctx.voice_client.is_paused():
            await self.call_event(
                "on_music_error",
                ctx,
                NotPlaying("Client is not playing anything currently"),
            )

        return now_playing

    @ensure_connection(check_playing=True, check_queue=True)
    async def queueloop(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Toggles the queue loop.

        :param ctx: The context
        :type ctx: commands.Context
        :return: A bool indicating if the queue loop is now enabled or disabled.
        :rtype: Optional[bool]
        """

        queue = self.queue[ctx.guild.id]

        queue.loop = (
            Loops.QUEUE_LOOP
            if self.queue[ctx.guild.id].loop != Loops.QUEUE_LOOP
            else Loops.NO_LOOP
        )

        if queue.loop == Loops.QUEUE_LOOP:
            queue.queue_loop_start = queue.pos

        return queue.loop == Loops.QUEUE_LOOP

    @ensure_connection(check_playing=True, check_queue=True)
    async def shuffle(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Toggles the shuffle feature.

        :param commands.Context ctx: The context
        :return: A bool indicating if the queue loop is now enabled or disabled.
        :rtype: Optional[bool]
        """

        queue = self.queue[ctx.guild.id]

        queue.shuffle = not queue.shuffle
        if queue.shuffle:
            queue.original_queue = queue.queue

            play_queue = queue.queue[queue.pos + 1 :]
            shuffled_queue = random.sample(play_queue, len(play_queue))
            queue.queue = (
                queue.queue[: queue.pos] + [queue.now_playing] + shuffled_queue
            )

        return queue.shuffle

    @ensure_connection(check_playing=True, check_queue=True)
    async def autoplay(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Toggles the autoplay feature.

        :param commands.Context ctx: The context
        :return: A bool indicating if autoplay is now enabled or disabled.
        :rtype: Optional[bool]
        """

        self.queue[ctx.guild.id].autoplay = not self.queue[ctx.guild.id].autoplay
        return self.queue[ctx.guild.id].autoplay

    @ensure_connection(check_playing=True, check_queue=True)
    async def loop(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Toggles the loop.

        :param ctx: The context
        :type ctx: commands.Context
        :return: A bool indicating if the loop is now enabled or disabled.
        :rtype: Optional[bool]
        """

        self.queue[ctx.guild.id].loop = (
            Loops.LOOP if self.queue[ctx.guild.id].loop != Loops.LOOP else Loops.NO_LOOP
        )
        return self.queue[ctx.guild.id].loop == Loops.LOOP

    @ensure_connection(check_queue=True)
    async def get_queue(self, ctx: commands.Context) -> Optional[QueueManager]:
        """
        |coro|

        Returns the queue of ctx.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: The queue.
        :rtype: Optional[QueueManager]
        """

        return self.queue[ctx.guild.id]

    @staticmethod
    def parse_duration(duration: Union[str, float], hour_format: bool = True) -> str:
        """
        |coro|

        Returns parsed duration.

        :param bool hour_format: A bool indicating if the parse should contain hours.
        :param duration: The duration.
        :type duration: Union[str, float]
        :param duration: Format Hours.
        :type duration: bool
        :return: The parsed duration.
        :rtype: str
        """

        if duration == "LIVE":
            return duration

        time_format = "%H:%M:%S" if hour_format else "%M:%S"

        return time.strftime(time_format, time.gmtime(round(duration)))

    @ensure_connection(check_queue=True)
    async def move(
        self, ctx: commands.Context, player_index: int, new_index: int
    ) -> Optional[Player]:
        """

        :param player_index: The index of the player that you want to move.
        :param new_index: The index you want to move the player to.
        :param ctx: Context to fetch the queue from
        :return: The player object that was moved
        :rtype: Optional[Player]
        """

        queue = await self.get_queue(ctx)
        player_index += queue.pos
        new_index += queue.pos

        if new_index > len(queue.queue) or player_index > len(queue.queue):
            return await self.call_event(
                "on_music_error", ctx, InvalidSkipIndex("Skip index is invalid")
            )

        player = queue.remove(player_index)
        queue.queue.insert(new_index, player)
        return player



from enum import Enum


__all__ = ("Loops", "PlaylistType", "ManagerType")


class Loops(Enum):
    NO_LOOP = 0
    LOOP = 1
    QUEUE_LOOP = 2


class PlaylistType(Enum):
    SPOTIFY = 0
    YOUTUBE = 1


class ManagerType(Enum):
    FFMPEG = 0
    LAVALINK = 1



""""
If InviteTracker is used in any way that breaks disnake TOS we, (the disnakeSuperUtils team)
are not responsible or liable in any way.
InviteTracker by disnakeSuperUtils was not intended to violate Discord TOS in any way.
In case we are contacted by Discord, we will remove any and all features that violate the Discord ToS.
Please feel free to read the Discord Terms of Service https://discord.com/terms.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Union, Optional
import disnake

from .base import DatabaseChecker

if TYPE_CHECKING:
    from disnake.ext import commands


@dataclass
class InviteAccount:
    """
    Represents an InviteAccount.
    """

    invite_tracker: InviteTracker
    member: disnake.Member

    async def get_invited_users(self):
        return await self.invite_tracker.get_members_invited(
            self.member, self.member.guild
        )


class InviteTracker(DatabaseChecker):
    def __init__(self, bot: commands.Bot):
        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "member": "snowflake",
                    "members_invited": "string",
                }
            ],
            ["invites"],
        )
        self.bot = bot
        self.cache = {}

        self.bot.loop.create_task(self.__initialize_cache())

        self.bot.add_listener(self.__cleanup_guild_cache, "on_guild_remove")
        self.bot.add_listener(self.__update_guild_cache, "on_guild_add")
        self.bot.add_listener(self.__track_invite, "on_invite_create")
        self.bot.add_listener(self.__cleanup_invite, "on_invite_delete")

    async def get_invite(self, member: disnake.Member) -> Optional[disnake.Invite]:
        for inv in await member.guild.invites():
            for invite in self.cache[member.guild.id]:
                if invite.revoked:
                    self.cache[invite.guild.id].remove(invite)
                    return

                if invite.code == inv.code and inv.uses - invite.uses == 1:
                    await self.__update_guild_cache(member.guild)
                    return inv

    @DatabaseChecker.uses_database
    async def get_members_invited(
        self, user: Union[disnake.User, disnake.Member], guild: disnake.Guild
    ):
        invited_members = await self.database.select(
            self.tables["invites"],
            ["members_invited"],
            {"guild": guild.id, "member": user.id},
        )

        if not invited_members:
            return []

        invited_members = invited_members["members_invited"]
        if isinstance(invited_members, str):
            return [
                int(invited_member)
                for invited_member in invited_members.split("\0")
                if invited_member
            ]

    async def fetch_inviter(
        self, invite: disnake.Invite
    ) -> Union[disnake.Member, disnake.User]:
        inviter = invite.guild.get_member(invite.inviter.id)
        return inviter if inviter else await self.bot.get_user(invite.inviter.id)

    @DatabaseChecker.uses_database
    async def register_invite(
        self,
        invite: disnake.Invite,
        member: disnake.Member,
        inviter: Union[disnake.Member, disnake.User],
    ) -> None:
        invited_members = await self.get_members_invited(inviter, invite.guild)
        if member.id in invited_members:
            return

        invited_members.append(member.id)
        invited_members_sql = "\0".join(
            str(invited_member) for invited_member in invited_members
        )

        await self.database.updateorinsert(
            self.tables["invites"],
            {"members_invited": invited_members_sql},
            {"guild": invite.guild.id, "member": inviter.id},
            {
                "guild": invite.guild.id,
                "member": inviter.id,
                "members_invited": invited_members_sql,
            },
        )

    async def __initialize_cache(self) -> None:
        await self.bot.wait_until_ready()

        for guild in self.bot.guilds:
            try:
                self.cache[guild.id] = await guild.invites()
            except disnake.Forbidden:
                pass

    async def __update_guild_cache(self, guild: disnake.Guild) -> None:
        try:
            self.cache[guild.id] = await guild.invites()
        except disnake.Forbidden:
            pass

    async def __track_invite(self, invite: disnake.Invite) -> None:
        self.cache[invite.guild.id].append(invite)

    async def __cleanup_invite(self, invite: disnake.Invite) -> None:
        if invite in self.cache[invite.guild.id]:
            self.cache[invite.guild.id].remove(invite)

    async def __cleanup_guild_cache(self, guild: disnake.Guild) -> None:
        self.cache.pop(guild.id)

    @DatabaseChecker.uses_database
    def get_user_info(self, member: disnake.Member) -> InviteAccount:
        return InviteAccount(self, member)
